// Game:           Prince of Persia (SNES)
// Category:       100%
// Emulator:       snes9x-1.60-win32-x64
// Original by:    David from forum.princed.org
// Updated:        2023-01-21
// IGT timing:     YES

// Base memory path:
// snes9x-1.60-win32-x64: "snes9x-x64.exe"+008D8C38

// Known bugs:
// The timer will not start if the player fast-forwards through START GAME.

// Open questions:
// * What if the player (dies and) restarts a level?

state("snes9x-x64")
{
	byte720 LevelObjects          : "snes9x-x64.exe", 0x008D8C38, 0x1E170; // 7F:E170; 24 rooms * 3 rows * 10 columns
	byte NextLevel                : "snes9x-x64.exe", 0x008D8C38, 0x0544; // 7E:0544; 0 means LEVEL 1, 1 means LEVEL 2, and so on.
	byte CurrentLevel             : "snes9x-x64.exe", 0x008D8C38, 0x0579; // 7E:0579; 0 means LEVEL 1, 1 means LEVEL 2, and so on.
	//byte CharScrn                 : "snes9x-x64.exe", 0x008D8C38, 0x0462; // 7E:0462
	byte KidMaxHP                 : "snes9x-x64.exe", 0x008D8C38, 0x050A; // 7E:050A
	ushort FrameCount             : "snes9x-x64.exe", 0x008D8C38, 0x052D; // 7E:052D
	byte FrameCountHalf           : "snes9x-x64.exe", 0x008D8C38, 0x052F; // 7E:052F
	byte StartRoom                : "snes9x-x64.exe", 0x008D8C38, 0x1F110; // 7F:F110
	byte Kid_room                 : "snes9x-x64.exe", 0x008D8C38, 0x0472; // 7E:0472

	sbyte Guard_alive             : "snes9x-x64.exe", 0x008D8C38, 0x0486; // 7E:0486; < 0 if alive, >= 0 if dead
	byte Guard_room               : "snes9x-x64.exe", 0x008D8C38, 0x0482; // 7E:0482
	byte Guard_direction          : "snes9x-x64.exe", 0x008D8C38, 0x047A; // 7E:047A; 0x7F if there is no guard
	byte Guard_hp                 : "snes9x-x64.exe", 0x008D8C38, 0x050B; // 7E:050B
	byte Guard_frame              : "snes9x-x64.exe", 0x008D8C38, 0x0477; // 7E:0477
	ushort Guard_sequence_pointer : "snes9x-x64.exe", 0x008D8C38, 0x0480; // 7E:0480
}

startup
{
	//refreshRate = 30; // Prince of Persia runs at 12 fps, let's update 2.5x as often to be sure

	settings.Add("test_skip", false, "For testing: Quickly skip over each action if the game is past its level.");
	settings.Add("any_order", true, "Allow doing the actions in any order.");

	// This is the list of actions needed for 100% completion.
	// Keep this in sync with the splits!
	vars.checks = new int[][]
	{
		// columns: type, level, room, tile, done
		// types:
		// 'p' = pick up (potion, sword)
		// 'g' = guard (ignores the tile)
		// 'x' = level exit (ignores the room and the tile)
		// "done" starts off as 0 and is set to 1 when the action is done,
		// This column is there to support completing actions in any order.

		// Level 1
		new int[]{'p', 1,  8, 27, 0}, // Level 1: Heal potion at room 8, tile 27 (attic, under the guard)
		new int[]{'p', 1, 12,  4, 0}, // Level 1: Heal potion at room 12, tile 4 (cellar)
		new int[]{'p', 1, 20, 25, 0}, // Level 1: Sword at room 20, tile 25
		new int[]{'p', 1, 23,  1, 0}, // Level 1: Heal potion at room 23, tile 1 (attic left, above the skeleton)
		new int[]{'g', 1,  8, -1, 0}, // Level 1: Guard at room 8, tile 6 (attic)
		new int[]{'g', 1,  1, -1, 0}, // Level 1: Guard at room 1, tile 17 (before the exit)
		new int[]{'x', 1, -1, -1, 0}, // Level 1: Level exit

		// Level 2
		new int[]{'g', 2,  7, -1, 0}, // Level 2: Guard at room 7
		new int[]{'p', 2,  8, 14, 0}, // Level 2: Heal potion at room 2, tile 14 (room below the guard)
		new int[]{'g', 2, 11, -1, 0}, // Level 2: Guard at room 11 (still underground)
		new int[]{'p', 2,  5, 26, 0}, // Level 2: Heal potion at room 5, tile 26 (above the ceiling)
		new int[]{'g', 2,  2, -1, 0}, // Level 2: Guard at room 2 (at the skeleton)
		new int[]{'p', 2,  1, 27, 0}, // Level 2: Heal potion at room 1, tile 27 (between the gates)
		new int[]{'g', 2,  0, -1, 0}, // Level 2: Guard at room 0 (bluish, at the spikes)
		new int[]{'p', 2, 13, 28, 0}, // Level 2: Life potion at room 13, tile 28
		new int[]{'p', 2, 23, 25, 0}, // Level 2: Heal potion at room 23, tile 25 (ceiling)
		new int[]{'g', 2, 16, -1, 0}, // Level 2: Guard at room 16 (bluish)
		new int[]{'g', 2, 15, -1, 0}, // Level 2: Guard at room 15 (next to the exit)
		new int[]{'g', 2, 12, -1, 0}, // Level 2: Guard at room 12 (bluish) (under the button which opens the exit)
		new int[]{'x', 2, -1, -1, 0}, // Level 2: Level exit

		// Level 3
		new int[]{'g', 3, 21, -1, 0}, // Level 3: Guard at room 21
		new int[]{'g', 3,  9, -1, 0}, // Level 3: Brown Skeleton at room 9
		new int[]{'p', 3,  9, 27, 0}, // Level 3: Life potion at room 9, tile 27 (at the brown skeleton)
		new int[]{'p', 3,  8,  2, 0}, // Level 3: Life potion at room 8, tile 2 (behind chompers)
		new int[]{'p', 3, 10,  8, 0}, // Level 3: Heal potion at room 10, tile 8 (after the first gate)
		new int[]{'p', 3, 18, 11, 0}, // Level 3: Hurt potion at room 18, tile 11 (on the way to the exit opener)
		new int[]{'p', 3, 18, 21, 0}, // Level 3: Heal potion at room 18, tile 21
		// These don't work yet, even if I disable the Guard_room check.
		//new int[]{'g', 3, 13, -1, 0}, // Level 3: Skeleton falls out of room 13
		//new int[]{'g', 3, 14, -1, 0}, // Level 3: Skeleton falls out of room 14
		new int[]{'p', 3, 12, 16, 0}, // Level 3: Heal potion at room 12, tile 16 (underground)
		new int[]{'g', 3, 20, -1, 0}, // Level 3: Skeleton crushed in room 20
		new int[]{'x', 3, -1, -1, 0}, // Level 3: Level exit

		// Level 4
		new int[]{'g', 4,  8, -1, 0}, // Level 4: Guard at room 8
		new int[]{'p', 4, 13, 12, 0}, // Level 4: Heal potion at room 13, tile 12 (behind the guard)
		new int[]{'p', 4, 12, 22, 0}, // Level 4: Heal potion at room 12, tile 22 (ceiling)
		new int[]{'g', 4,  2, -1, 0}, // Level 4: Guard at room 2
		new int[]{'g', 4,  3, -1, 0}, // Level 4: Guard at room 3
		new int[]{'p', 4, 17, 15, 0}, // Level 4: Heal potion at room 17, tile 15 (ceiling)
		new int[]{'p', 4, 17, 17, 0}, // Level 4: Heal potion at room 17, tile 17 (ceiling)
		new int[]{'g', 4, 16, -1, 0}, // Level 4: Guard at room 16
		new int[]{'p', 4, 22, 16, 0}, // Level 4: Life potion at room 22, tile 16
		new int[]{'x', 4, -1, -1, 0}, // Level 4: Level exit

		// TODO: Levels 5-20
	};

	vars.check_index = 0;
	vars.playing = false;
	vars.reset_guards = (Action)delegate()
	{
		int[] array = new int[24];
		for (int i = 0; i < 24; i++) array[i] = i;
		vars.guard_original_rooms = array;
	};
	vars.reset_actions = (Action)delegate()
	{
		// Mark all actions as not done.
		for (int check_index = 0; check_index < vars.checks.Length; check_index++)
		{
			vars.checks[check_index][4] = 0;
		}
	};
	vars.reset_guards();
	// Reset if this autosplitter is (changed and) reloaded, because then we lose track of the current split anyway.
	vars.reset = true;
}

reset
{
	// KidMaxHP == 0 when the game is reset.
	// CurrentLevel == 0 on level 1 *and* in the main menu!
	// CurrentLevel >= 21 are the demo and training levels.
	// CurrentLevel == 128 during cutscenes.
	// CurrentLevel == 252 after the demo level was interrupted and before the main menu fully appears.
	// CurrentLevel == 254 after GAME END was selected and before the main menu fully appears.
	// CurrentLevel == 255 when restarting the level.
	bool reset = (
		current.KidMaxHP == 0
		|| (current.CurrentLevel >= 21 && current.CurrentLevel < 128)
		|| current.CurrentLevel == 252
		|| current.CurrentLevel == 254
		|| vars.reset
	);
	if (reset)
	{
		print("reset");
		print("CurrentLevel = " + current.CurrentLevel);
		vars.check_index = 0;
		vars.playing = false;
		vars.reset_guards();
		vars.reset_actions();
		vars.reset = false;
	}
	return reset;
}

start
{
	// This should become true only when the player selects START GAME.
	// We check StartRoom to prevent false positives, as otherwise the condition is true after the player quits the demo/training level.
	// PROBLEM: This doesn't work if I fast-forward through START GAME, because then FrameCount is 0 only for a too short time.
	bool start = (
		current.KidMaxHP == 3
		&& current.FrameCount == 0
		&& current.CurrentLevel == 0
		&& current.StartRoom == 22
		//&& current.StartRoom == current.Kid_room
	);
	if (start)
	{
		print("start");
		vars.check_index = 0;
		vars.playing = true;
		vars.reset_guards();
		vars.reset_actions();
	}
	return start;
}

update
{
	if (!vars.playing) return;

	//print("settings: any_order = " + settings["any_order"]);
	//print("settings: test_skip = " + settings["test_skip"]);
	//print("vars.check_index = " + vars.check_index);

	// CurrentLevel == 255 when restarting the level.
	if (current.CurrentLevel == 255) vars.reset_guards();

	// CurrentLevel == 128 during cutscenes.
	if (current.CurrentLevel >= 128) return;

	//print("It's alive!");
	//print("CurrentLevel = "+current.CurrentLevel);
	//print("NextLevel = "+current.NextLevel);
	//print("KidMaxHP = "+current.KidMaxHP);

	// If a guard moved to another room:
	if (current.Guard_direction != 0x7F
		&& old.Guard_direction != 0x7F
		&& current.Guard_room != old.Guard_room
		&& current.Guard_direction == old.Guard_direction
		&& current.Guard_hp == old.Guard_hp
		// TODO: Check more attributes of the guard to make sure they are the same guard?
		// TODO: Check if the prince is in fight mode? Or, can guards move between rooms even if he isn't?
		)
	{
		// If there is a guard in room R, then guard_original_rooms[R] tells the room where that guard started the level.
		int guard_original_room = vars.guard_original_rooms[old.Guard_room];
		print("A guard moved from room "+old.Guard_room+" to "+current.Guard_room+". (original room: "+guard_original_room+")");
		vars.guard_original_rooms[current.Guard_room] = guard_original_room;
	}

	// Stop if the player already completed all actions.
	if (vars.check_index >= vars.checks.Length) return;

	// DONE: Allow doing the actions in any order within a level.
	int first_index, last_index;
	if (settings["any_order"])
	{
		// Look at all actions.
		// Ideally we would only look at the action of the current level...
		first_index = vars.check_index;
		last_index = vars.checks.Length - 1;
	}
	else
	{
		// Look at only the current action.
		first_index = last_index = vars.check_index;
	}

	for (int check_index = first_index; check_index <= last_index; check_index++)
	{
		// If a check matches, mark it as done (in a new check[4]).
		// If vars.checks[vars.check_index] is done then mark a split.

		//int[] check = vars.checks[vars.check_index];
		int[] check = vars.checks[check_index];
		char type = (char)check[0];
		int level = check[1];
		int room = check[2];
		int tile = check[3];
		//print("check_index = "+vars.check_index);
		//print("type = "+type);
		//print("level = "+level);
		//print("room = "+room);
		//print("tile = "+tile);

		if (settings["test_skip"])
		{
			// For testing: Quickly skip over each action if the game is past its level.
			// DONE: Added a setting for this.
			if (current.CurrentLevel+1 > level)
			{
				check[4] = 1;
				vars.reset_guards();
				continue;
			}
		}

		// Type IDs of objects which can be picked up:
		// (Source: https://www.princed.org/wiki/SNES_format#Levels)
		// 0x0A = sword
		// 0x12 = heal potion
		// 0x13 = life potion
		// 0x14 = hurt potion
		// 0x15 = upside down potion
		// 0x16 = slow fall potion
		// 0x1A = potion of warp
		// 0x1C = kill potion

		if (current.CurrentLevel+1 == level)
		{

			switch (type)
			{
				case 'p': // pick up
					if (current.CurrentLevel+1 == level)
					{
						int offset = room*30+tile;
						// TODO: Compare the old value against the specific object type?
						// SOLVED: If we check the old value in any way, then we won't detect actions if the player does them out of order.
						// PROBLEM: If we *don't* check the old value, we can get false positives before the level is fully loaded.
						if (current.LevelObjects[offset] == 0x00 && old.LevelObjects[offset] != 0x00)
						{
							print("The player picked up the item at: level "+level+", room "+room+", tile "+tile);
							check[4] = 1;
						}
					}
				break;

				case 'g': // guard
					if (
						(
							// the guard died onscreen
							current.Guard_direction != 0x7F // a guard is present
							&& (
								(current.Guard_alive == 0 && old.Guard_alive != 0) // died
								// We need to check these separately because skeletons are not marked as dead in Guard_alive.
								|| (current.Guard_frame == 184 && old.Guard_frame != 184) // skeleton crushed
								|| (old.Guard_frame == 177 && current.Guard_frame == 186) // brown skeleton collapsed
							)
						)
						||
						(
							// the guard fell out of the room
							(current.Guard_direction == 0x7F && old.Guard_direction != 0x7F) // the guard disappeared
							&& (current.Guard_room == old.Guard_room) // PROBLEM: This is false for the level 3 skeleton.
							&& (current.Kid_room == old.Kid_room) // not because the kid moved to another room
						)
					)
					{
						int guard_original_room = vars.guard_original_rooms[current.Guard_room];
						print("A guard died in room " + current.Guard_room + " of level " + (current.CurrentLevel+1)
							+ ". (original room: "+guard_original_room+")");
						if (current.CurrentLevel+1 == level && /*current.Guard_room*/ guard_original_room == room)
						{
							check[4] = 1;
						}
					}
				break;

				case 'x': // level exit
					if (current.CurrentLevel+1 == level
						&& current.NextLevel != current.CurrentLevel)
					{
						check[4] = 1;
						vars.reset_guards();
					}
				break;
			}

		}

	}

}

split
{
	int[] check = vars.checks[vars.check_index];
	bool split = (check[4] == 1);
	if (split) vars.check_index++;
	return split;
}

gameTime
{
	// 1 minute = 0x01A9 frames, 1 second = 7 frames.
	// So 1 minute is slightly more than 60 seconds, and they don't even match real time.
	return TimeSpan.FromMinutes( ((double)current.FrameCount + 0.5 * (double)current.FrameCountHalf) / (double)0x01A9);
}

isLoading
{
	// Force Game Time to show the actual Game Time...
	return true;
}
